import math
from enum import Enum
from typing import List, Optional, Sequence, Tuple, Type

from redis.commands.core import Script

from ..constants import RateLimiterType, StoreType
from ..store import BaseAtomicAction, MemoryStoreBackend, RedisStoreBackend
from ..types import AtomicActionTypeT, KeyT, RateLimiterTypeT, StoreValueT
from ..utils import now_ms, now_sec
from . import BaseRateLimiter, RateLimitResult, RateLimitState


class SlidingWindowAtomicActionType(Enum):
    """Enumeration for types of AtomicActions used in SlidingWindowRateLimiter."""

    LIMIT: AtomicActionTypeT = "limit"


class RedisLimitAtomicAction(BaseAtomicAction):
    """Redis-based implementation of AtomicAction for SlidingWindowRateLimiter."""

    TYPE: AtomicActionTypeT = SlidingWindowAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.REDIS.value

    SCRIPTS: str = """
    local period = tonumber(ARGV[1])
    local limit = tonumber(ARGV[2])
    local cost = tonumber(ARGV[3])
    local now_ms = tonumber(ARGV[4])

    local current = redis.call("INCRBY", KEYS[1], cost)
    if current == cost then
        redis.call("EXPIRE", KEYS[1], 3 * period)
    end

    local previous = redis.call("GET", KEYS[2])
    if previous == false then
        previous = 0
    end

    local period_ms = period * 1000
    local current_proportion = (now_ms % period_ms) / period_ms
    previous = math.floor((1 - current_proportion) * previous)
    local used = previous + current

    local retry_after = 0
    local limited = used > limit
    if limited then
        if cost <= previous then
            retry_after = (1 - current_proportion) * period * cost / previous
        else
            retry_after = (1- current_proportion) * period
        end
    end

    return {limited, used, tostring(retry_after)}
    """

    def __init__(self, backend: RedisStoreBackend):
        self._script: Script = backend.get_client().register_script(self.SCRIPTS)

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int, float]:
        limited, used, retry_after = self._script(keys, args)
        return limited, used, float(retry_after)


class MemoryLimitAtomicAction(BaseAtomicAction):
    """Memory-based implementation of AtomicAction for SlidingWindowRateLimiter."""

    TYPE: AtomicActionTypeT = SlidingWindowAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.MEMORY.value

    def __init__(self, backend: MemoryStoreBackend):
        self._backend: MemoryStoreBackend = backend

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int, float]:
        with self._backend.lock:
            current_key: str = keys[0]
            previous_key: str = keys[1]
            period: int = args[0]
            limit: int = args[1]
            cost: int = args[2]

            current: Optional[int] = self._backend.get(current_key)
            if current is None:
                current = cost
                self._backend.set(current_key, cost, 3 * period)
            else:
                current += cost
                self._backend.get_client()[current_key] = current

            period_ms: int = period * 1000
            current_proportion: float = (args[3] % period_ms) / period_ms
            previous: int = math.floor(
                (1 - current_proportion) * (self._backend.get(previous_key) or 0)
            )

            used: int = previous + current
            limited: int = (0, 1)[used > limit]
            if limited:
                if cost <= previous:
                    retry_after = (1 - current_proportion) * period * cost / previous
                else:
                    retry_after = (1 - current_proportion) * period
            else:
                retry_after = 0

            return limited, used, retry_after


class SlidingWindowRateLimiter(BaseRateLimiter):
    """Concrete implementation of BaseRateLimiter using sliding window as algorithm."""

    class Meta:
        type: RateLimiterTypeT = RateLimiterType.SLIDING_WINDOW.value

    @classmethod
    def _default_atomic_action_classes(cls) -> List[Type[BaseAtomicAction]]:
        return [RedisLimitAtomicAction, MemoryLimitAtomicAction]

    @classmethod
    def _supported_atomic_action_types(cls) -> List[AtomicActionTypeT]:
        return [SlidingWindowAtomicActionType.LIMIT.value]

    def _prepare(self, key: str) -> Tuple[str, str, int, int]:
        period: int = self.quota.get_period_sec()
        current_idx: int = now_sec() // period
        current_key: str = self._prepare_key(f"{key}:period:{current_idx}")
        previous_key: str = self._prepare_key(f"{key}:period:{current_idx - 1}")
        return current_key, previous_key, period, self.quota.get_limit()

    def _limit(self, key: str, cost: int = 1) -> RateLimitResult:
        current_key, previous_key, period, limit = self._prepare(key)
        limited, used, retry_after = self._atomic_actions[
            SlidingWindowAtomicActionType.LIMIT.value
        ].do([current_key, previous_key], [period, limit, cost, now_ms()])
        return RateLimitResult(
            limited=bool(limited),
            state=RateLimitState(
                limit=limit,
                remaining=max(0, limit - used),
                reset_after=period,
                retry_after=retry_after,
            ),
        )

    def _peek(self, key: str) -> RateLimitState:
        current_key, previous_key, period, limit = self._prepare(key)
        period_ms: int = period * 1000
        current_proportion: float = (now_ms() % period_ms) / period_ms
        previous: int = math.floor(
            (1 - current_proportion) * (self._store.get(previous_key) or 0)
        )
        used: int = previous + (self._store.get(current_key) or 0)

        return RateLimitState(
            limit=limit, remaining=max(0, limit - used), reset_after=period
        )



from typing import Dict, Union

_StringLikeT = str
_NumberLikeT = Union[int, float]

KeyT = _StringLikeT
StoreValueT = _NumberLikeT
StoreDictValueT = Dict[KeyT, _NumberLikeT]
StoreBucketValueT = Union[_NumberLikeT, StoreDictValueT]

AtomicActionTypeT = str

RateLimiterTypeT = str



import threading
import time
from types import TracebackType
from typing import Callable, Optional, Type

from .constants import RateLimiterType
from .exceptions import DataError, LimitedError
from .rate_limiter import (
    BaseRateLimiter,
    Quota,
    RateLimiterRegistry,
    RateLimitResult,
    RateLimitState,
    per_min,
)
from .store import BaseStore, MemoryStore
from .types import KeyT, RateLimiterTypeT
from .utils import now_sec_f


class Throttled:
    # Non-blocking mode constant
    _NON_BLOCKING: float = -1
    # Interval between retries in seconds
    _WAIT_INTERVAL: float = 0.5
    # Minimum interval between retries in seconds
    _WAIT_MIN_INTERVAL: float = 0.2

    def __init__(
        self,
        key: Optional[KeyT] = None,
        timeout: Optional[float] = None,
        using: Optional[RateLimiterTypeT] = None,
        quota: Optional[Quota] = None,
        store: Optional[BaseStore] = None,
    ):
        """Initializes the Throttled class.
        :param key: The unique identifier for the rate limit subject.
                    eg: user ID or IP address.
        :param timeout: Maximum wait time in seconds when rate limit is
                        exceeded.
                        (Default) If set to -1, it will return immediately.
                        Otherwise, it will block until the request can be
                        processed or the timeout is reached.
        :param using: The type of rate limiter to use, default: token_bucket.
        :param quota: The quota for the rate limiter, default: 60 requests per minute.
        :param store: The store to use for the rate limiter, default: MemoryStore.
        """
        # TODO Support key prefix.
        # TODO Support extract key from params.
        # TODO Support get cost weight by key.
        self.key: Optional[str] = key

        if timeout is None:
            timeout = self._NON_BLOCKING
        self.timeout: float = timeout
        self._validate_timeout(self.timeout)

        self._quota: Quota = quota or per_min(60)
        self._store: BaseStore = store or MemoryStore()
        self._limiter_cls: Type[BaseRateLimiter] = RateLimiterRegistry.get(
            using or RateLimiterType.TOKEN_BUCKET.value
        )

        self._lock: threading.Lock = threading.Lock()
        self._limiter: Optional[BaseRateLimiter] = None

    def _get_limiter(self) -> BaseRateLimiter:
        """Lazily initializes and returns the rate limiter instance."""
        if self._limiter:
            return self._limiter

        with self._lock:
            # Double-check locking to ensure thread safety.
            if self._limiter:
                return self._limiter

            self._limiter = self._limiter_cls(self._quota, self._store)
            return self._limiter

    def __enter__(self) -> RateLimitResult:
        """Context manager to apply rate limiting to a block of code.
        :return: RateLimitResult
        :raise: LimitedError if rate limit is exceeded.
        """
        result: RateLimitResult = self.limit()
        if result.limited:
            raise LimitedError(rate_limit_result=result)
        return result

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        pass

    def __call__(self, func: Callable) -> Callable:
        """Decorator to apply rate limiting to a function."""

        if not self.key:
            raise DataError(f"Invalid key: {self.key}, must be a non-empty key.")

        def _inner(*args, **kwargs):
            # TODO Add options to ignore state.
            result: RateLimitResult = self.limit()
            if result.limited:
                raise LimitedError(rate_limit_result=result)
            return func(*args, **kwargs)

        return _inner

    @classmethod
    def _validate_cost(cls, cost: int) -> None:
        """Validate the cost of the current request.
        :param cost: The cost of the current request in terms of
                     how much of the rate limit quota it consumes.
        :raise: DataError if the cost is not a positive integer.
        """
        if isinstance(cost, int) and cost > 0:
            return

        raise DataError(
            f"Invalid cost: {cost}, must be an integer greater than 0.".format(cost=cost)
        )

    @classmethod
    def _validate_timeout(cls, timeout: float) -> None:
        """Validate the timeout value.
        :param timeout: Maximum wait time in seconds when rate limit is exceeded.
        :raise: DataError if the timeout is not a positive float or -1(non-blocking).
        """

        if timeout == cls._NON_BLOCKING:
            return

        if (isinstance(timeout, float) or isinstance(timeout, int)) and timeout > 0:
            return

        raise DataError(
            f"Invalid timeout: {timeout}, must be a positive float or -1(non-blocking)."
        )

    def _get_key(self, key: Optional[KeyT] = None) -> KeyT:
        # Use the provided key if available.
        if key:
            return key

        if self.key:
            return self.key

        raise DataError(f"Invalid key: {key}, must be a non-empty key.")

    def _get_timeout(self, timeout: Optional[float] = None) -> float:
        if timeout is not None:
            self._validate_timeout(timeout)
            return timeout

        return self.timeout

    def _wait(self, timeout: float, retry_after: float) -> None:
        """Wait for the specified timeout or until retry_after is reached."""
        if retry_after <= 0:
            return

        start_time: float = now_sec_f()
        while True:
            # WAIT_INTERVAL: Chunked waiting interval to avoid long blocking periods.
            # Also helps reduce actual wait time considering thread context switches.
            # WAIT_MIN_INTERVAL: Minimum wait interval to prevent busy-waiting.
            sleep_time: float = max(
                min(retry_after, self._WAIT_INTERVAL), self._WAIT_MIN_INTERVAL
            )

            # Sleep for the specified time.
            time.sleep(sleep_time)

            # Calculate the elapsed time since the start time.
            # Due to additional context switching overhead in multithread contexts,
            # we don't directly use sleep_time to calculate elapsed time.
            # Instead, we re-fetch the current time and subtract it from the start time.
            elapsed: float = now_sec_f() - start_time
            if elapsed >= retry_after or elapsed >= timeout:
                break

    def limit(
        self, key: Optional[KeyT] = None, cost: int = 1, timeout: Optional[float] = None
    ) -> RateLimitResult:
        """Apply rate limiting logic to a given key with a specified cost.
        :param key: The unique identifier for the rate limit subject.
                    eg: user ID or IP address.
                    Overrides the instance key if provided.
        :param cost: The cost of the current request in terms of how much
                     of the rate limit quota it consumes.
        :param timeout: Maximum wait time in seconds when rate limit is
                        exceeded.
                        If set to -1, it will return immediately.
                        Otherwise, it will block until the request can
                        be processed or the timeout is reached.
                        Overrides the instance timeout if provided.
        :return: RateLimitResult: The result of the rate limiting check.
        :raise: DataError if invalid parameters.
        """
        self._validate_cost(cost)

        key: KeyT = self._get_key(key)
        timeout: float = self._get_timeout(timeout)
        result: RateLimitResult = self._get_limiter().limit(key, cost)
        if timeout == self._NON_BLOCKING or not result.limited:
            return result

        # TODO: When cost > limit, return early instead of waiting.
        start_time: float = now_sec_f()
        while True:
            if result.state.retry_after > timeout:
                break

            self._wait(timeout, result.state.retry_after)

            result = self._get_limiter().limit(key, cost)
            if not result.limited:
                break

            elapsed: float = now_sec_f() - start_time
            if elapsed >= timeout:
                break

        return result

    def peek(self, key: KeyT) -> RateLimitState:
        """Retrieve the current state of rate limiter for the given key
           without actually modifying the state.
        :param key: The unique identifier for the rate limit subject.
                    eg: user ID or IP address.
        :return: RateLimitState - Representing the current state of
                 the rate limiter for the given key.
        """
        return self._get_limiter().peek(key)



import abc
import logging
from dataclasses import dataclass
from datetime import timedelta
from typing import Dict, List, Optional, Set, Type

from ..exceptions import SetUpError
from ..store import BaseAtomicAction, BaseStore
from ..types import AtomicActionTypeT, RateLimiterTypeT

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class Rate:
    """Rate represents the rate limit configuration."""

    # The time period for which the rate limit applies.
    period: timedelta

    # The maximum number of requests allowed within the specified period.
    limit: int


@dataclass
class Quota:
    """Quota represents the quota limit configuration."""

    # The base rate limit configuration.
    rate: Rate

    # Optional burst capacity that allows exceeding the rate limit momentarily.
    # Default is 0, which means no burst capacity.
    burst: int = 0

    def get_period_sec(self) -> int:
        return int(self.rate.period.seconds)

    def get_limit(self) -> int:
        return self.rate.limit


def per_sec(limit: int, burst: Optional[int] = None) -> Quota:
    """Create a quota representing the maximum requests and burst per second."""
    if burst is None:
        burst = limit
    return Quota(Rate(period=timedelta(seconds=1), limit=limit), burst=burst)


def per_min(limit: int, burst: Optional[int] = None) -> Quota:
    """Create a quota representing the maximum requests and burst per minute."""
    if burst is None:
        burst = limit
    return Quota(Rate(period=timedelta(minutes=1), limit=limit), burst=burst)


def per_hour(limit: int, burst: Optional[int] = None) -> Quota:
    """Create a quota representing the maximum requests and burst per hour."""
    if burst is None:
        burst = limit
    return Quota(Rate(period=timedelta(hours=1), limit=limit), burst=burst)


def per_day(limit: int, burst: Optional[int] = None) -> Quota:
    """Create a quota representing the maximum requests and burst per day."""
    return Quota(Rate(period=timedelta(days=1), limit=limit), burst=burst)


@dataclass
class RateLimitState:
    """RateLimitState represents the current state of the rate limiter for the given
    key."""

    # Limit represents the maximum number of requests allowed to pass in the initial
    # state.
    limit: int

    # Remaining represents the maximum number of requests allowed to pass for the given
    # key in the current state.
    remaining: int

    # ResetAfter represents the time in seconds for the RateLimiter to return to its
    # initial state. In the initial state, Limit=Remaining.
    reset_after: float

    # RetryAfter represents the time in seconds for the request to be retried, 0 if
    # the request is allowed.
    retry_after: float = 0


@dataclass
class RateLimitResult:
    """RateLimitState represents the result after executing the RateLimiter for the
    given key."""

    # Limited represents whether this request is allowed to pass.
    limited: bool

    state: RateLimitState


class RateLimiterRegistry:
    _RATE_LIMITERS: Dict[RateLimiterTypeT, Type["BaseRateLimiter"]] = {}

    @classmethod
    def register(cls, new_cls):
        try:
            cls._RATE_LIMITERS[new_cls.Meta.type] = new_cls
        except AttributeError as e:
            raise SetUpError("failed to register RateLimiter: {}".format(e))

    @classmethod
    def get(cls, _type: RateLimiterTypeT) -> Type["BaseRateLimiter"]:
        try:
            return cls._RATE_LIMITERS[_type]
        except KeyError:
            raise SetUpError("{} not found".format(_type))


class RateLimiterMeta(abc.ABCMeta):
    def __new__(cls, name, bases, attrs):
        new_cls = super().__new__(cls, name, bases, attrs)
        if not [b for b in bases if isinstance(b, RateLimiterMeta)]:
            return new_cls

        RateLimiterRegistry.register(new_cls)
        return new_cls


class BaseRateLimiter(metaclass=RateLimiterMeta):
    """Base class for RateLimiter."""

    KEY_PREFIX: str = "throttled:v1:"

    class Meta:
        type: RateLimiterTypeT = ""

    def __init__(
        self,
        quota: Quota,
        store: BaseStore,
        additional_atomic_actions: Optional[List[Type[BaseAtomicAction]]] = None,
    ) -> None:
        self.quota: Quota = quota
        self._store: BaseStore = store
        self._atomic_actions: Dict[AtomicActionTypeT, BaseAtomicAction] = {}
        self._register_atomic_actions(additional_atomic_actions or [])

    @classmethod
    @abc.abstractmethod
    def _default_atomic_action_classes(cls) -> List[Type[BaseAtomicAction]]:
        """Define the default AtomicAction classes for RateLimiter."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def _supported_atomic_action_types(cls) -> List[AtomicActionTypeT]:
        """Define the supported AtomicAction types for RateLimiter."""
        raise NotImplementedError

    @abc.abstractmethod
    def _limit(self, key: str, cost: int) -> RateLimitResult:
        raise NotImplementedError

    @abc.abstractmethod
    def _peek(self, key: str) -> RateLimitState:
        raise NotImplementedError

    def try_register_atomic_action(self, action_cls: Type[BaseAtomicAction]) -> None:
        """Try to register AtomicAction class.
        :param action_cls: AtomicAction class to be registered.
        :raise: SetUpError
        """
        action_cls.match_or_raise(self._store.TYPE)
        self._atomic_actions[action_cls.TYPE] = self._store.make_atomic(action_cls)

    def _validate_registered_atomic_actions(self) -> None:
        """Validate that all required AtomicAction types have been registered.
        :raise: SetUpError
        """
        supported_types: Set[AtomicActionTypeT] = set(
            self._supported_atomic_action_types()
        )
        registered_types: Set[AtomicActionTypeT] = set(self._atomic_actions.keys())

        missing_types: Set[str] = supported_types - registered_types
        if missing_types:
            raise SetUpError(
                "Missing AtomicActionTypes: expected {expected} but missing "
                "{missing}.".format(expected=supported_types, missing=missing_types)
            )

    def _register_atomic_actions(self, classes: List[Type[BaseAtomicAction]]) -> None:
        """Register AtomicAction classes for default and additional classes."""
        for action_cls in self._default_atomic_action_classes() + classes:
            try:
                self.try_register_atomic_action(action_cls)
            except SetUpError as e:
                logger.debug(
                    "Failed to register AtomicAction class for %s: %s", action_cls, e
                )
                pass

        self._validate_registered_atomic_actions()

    def _prepare_key(self, key: str) -> str:
        """Prepare the key by adding the prefix.
        :param key: The unique identifier for the rate limit subject.
        :return: The formatted key with prefix.

        # Benchmarks(TokenBucket)
        # Python 3.13.1 (main, Mar 29 2025, 16:29:36) [Clang 15.0.0 (clang-1500.3.9.4)]
        # Implementation: CPython
        # OS: Darwin 23.6.0, Arch: arm64
        #
        # >> Redis baseline
        # command    -> set key value
        # serial     -> 🕒Latency: 0.0589 ms/op, 🚀Throughput: 16828 req/s (--)
        # concurrent -> 🕒Latency: 1.9032 ms/op, 💤Throughput: 16729 req/s (⬇️-0.59%)
        #
        # >> Before preparing key
        # serial     -> 🕒Latency: 0.0722 ms/op, 🚀Throughput: 13740 req/s (--)
        # concurrent -> 🕒Latency: 2.3197 ms/op, 🚀Throughput: 13742 req/s (⬆️0.01%)
        #
        # >> After preparing key
        # serial     -> 🕒Latency: 0.0724 ms/op, 🚀Throughput: 13712 req/s (--)
        # concurrent -> 🕒Latency: 2.3126 ms/op, 🚀Throughput: 13782 req/s (⬆️0.51%)
        """
        return f"{self.KEY_PREFIX}{self.Meta.type}:{key}"

    def limit(self, key: str, cost: int = 1) -> RateLimitResult:
        """Apply rate limiting logic to a given key with a specified cost.
        :param key: The unique identifier for the rate limit subject.
                    eg: user ID or IP address.
        :param cost: The cost of the current request in terms of how much of the rate
                     limit quota it consumes.
        :return: A tuple containing two elements:
                 - RateLimitResult: Representing the result after executing the
                                    RateLimiter for the given key.
                 - RateLimitState: Representing the current state of the RateLimiter for
                                   the given key.
        """
        return self._limit(key, cost)

    def peek(self, key: str) -> RateLimitState:
        """Retrieve the current state of rate limiter for the given key
           without actually modifying the state.
        :param key: The unique identifier for the rate limit subject.
                    eg: user ID or IP address.
        :return: RateLimitState - Representing the current state of the rate limiter
                                  for the given key.
        """
        return self._peek(key)



# Copyright (c) 2011-2016 Andrey Antukh <niwi@niwi.nz>
# Copyright (c) 2011 Sean Bleier
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS`` AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import abc
from typing import Any, Dict, List, Optional, Tuple, Type, Union
from urllib.parse import ParseResult, ParseResultBytes, parse_qs, urlparse

from redis import ConnectionPool, Redis, Sentinel
from redis.connection import DefaultParser, to_bool

from throttled.exceptions import SetUpError
from throttled.utils import import_string


class BaseConnectionFactory(abc.ABC):
    """Base connection factory."""

    _pools: Dict[str, ConnectionPool] = {}

    def __init__(self, options: Dict[str, Any]):
        pool_cls_path: str = options.get(
            "CONNECTION_POOL_CLASS", "redis.connection.ConnectionPool"
        )
        self.pool_cls: Type[ConnectionPool] = import_string(pool_cls_path)
        self.pool_cls_kwargs: Dict[str, Any] = options.get("CONNECTION_POOL_KWARGS", {})

        redis_client_cls_path: str = options.get(
            "REDIS_CLIENT_CLASS", "redis.client.Redis"
        )
        self.redis_client_cls: Type[Redis] = import_string(redis_client_cls_path)
        self.redis_client_cls_kwargs: Dict[str, Any] = options.get(
            "REDIS_CLIENT_KWARGS", {}
        )

        self.options: Dict[str, Any] = options

    @abc.abstractmethod
    def make_connection_params(self, url: str = None) -> Dict[str, Any]:
        """Given a main connection parameters, build a complete dict of
        connection parameters.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def connect(self, url: Optional[str] = None) -> Redis:
        """Given a basic connection parameters, return a new connection."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_connection(self, params: Dict[str, Any]) -> Redis:
        """Given a now preformatted params, return a new connection.
        The default implementation uses a cached pools for create new connection.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_or_create_connection_pool(self, params: Dict[str, Any]) -> ConnectionPool:
        """
        Given a connection parameters and return a new
        or cached connection pool for them.

        Implement this method if you want distinct
        connection pool instance caching behavior.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_connection_pool(self, params: Dict[str, Any]) -> ConnectionPool:
        """
        Given a connection parameters, return a new connection pool for them.
        Implement this method if you want a custom behavior on creating connection pool.
        """
        raise NotImplementedError


class ConnectionFactory(BaseConnectionFactory):
    """Store connection pool by backend options.
    _pools is a process-global, as otherwise _pools is cleared every time.
    """

    DEFAULT_URL: str = "redis://localhost:6379/0"

    def make_connection_params(self, url: Optional[str] = None) -> Dict[str, Any]:
        kwargs: Dict[str, Any] = {
            "url": url or self.DEFAULT_URL,
            "parser_class": self.get_parser_cls(),
        }
        password: Optional[str] = self.options.get("PASSWORD", None)
        if password:
            kwargs["password"] = password

        socket_timeout: Any = self.options.get("SOCKET_TIMEOUT", None)
        if socket_timeout:
            if not isinstance(socket_timeout, (int, float)):
                raise SetUpError("Socket timeout should be float or integer")
            kwargs["socket_timeout"] = socket_timeout

        socket_connect_timeout: Any = self.options.get("SOCKET_CONNECT_TIMEOUT", None)
        if socket_connect_timeout:
            if not isinstance(socket_connect_timeout, (int, float)):
                raise SetUpError("Socket connect timeout should be float or integer")
            kwargs["socket_connect_timeout"] = socket_connect_timeout

        return kwargs

    def connect(self, url: Optional[str] = None) -> Redis:
        params: Dict[str, Any] = self.make_connection_params(url)
        return self.get_connection(params)

    def get_connection(self, params) -> Redis:
        pool: ConnectionPool = self.get_or_create_connection_pool(params)
        return self.redis_client_cls(
            connection_pool=pool, **self.redis_client_cls_kwargs
        )

    def get_parser_cls(self) -> Type[DefaultParser]:
        parser_cls_path: Optional[str] = self.options.get("PARSER_CLASS", None)
        if parser_cls_path is None:
            return DefaultParser
        return import_string(parser_cls_path)

    def get_or_create_connection_pool(self, params: Dict[str, Any]) -> ConnectionPool:
        key: str = params["url"]
        if key not in self._pools:
            self._pools[key] = self.get_connection_pool(params)
        return self._pools[key]

    def get_connection_pool(self, params: Dict[str, Any]) -> ConnectionPool:
        cp_params: Dict[str, Any] = dict(params)
        cp_params.update(self.pool_cls_kwargs)
        pool: ConnectionPool = self.pool_cls.from_url(**cp_params)

        if pool.connection_kwargs.get("password", None) is None:
            pool.connection_kwargs["password"] = params.get("password", None)
            pool.reset()

        return pool


class SentinelConnectionFactory(ConnectionFactory):
    def __init__(self, options):
        # allow overriding the default SentinelConnectionPool class
        options.setdefault(
            "CONNECTION_POOL_CLASS", "redis.sentinel.SentinelConnectionPool"
        )
        super().__init__(options)

        sentinels: List[Tuple[str, Union[str, int]]] = options.get("SENTINELS")
        if not sentinels:
            raise SetUpError("SENTINELS must be provided as a list of (host, port).")

        # provide the connection pool kwargs to the sentinel in case it
        # needs to use the socket options for the sentinels themselves
        connection_kwargs = self.make_connection_params(None)
        connection_kwargs.pop("url")
        connection_kwargs.update(self.pool_cls_kwargs)
        self._sentinel: Sentinel = Sentinel(
            sentinels,
            sentinel_kwargs=options.get("SENTINEL_KWARGS"),
            **connection_kwargs,
        )

    def get_connection_pool(self, params: Dict[str, Any]) -> ConnectionPool:
        """Given a connection parameters, return a new sentinel connection
        pool for them."""
        url: Union[ParseResult, ParseResultBytes] = urlparse(params["url"])

        # explicitly set service_name and sentinel_manager for the
        # SentinelConnectionPool constructor since will be called by from_url.
        cp_params: Dict[str, Any] = dict(params)
        cp_params.update(service_name=url.hostname, sentinel_manager=self._sentinel)
        pool: ConnectionPool = super().get_connection_pool(cp_params)

        # convert "is_master" to a boolean if set on the URL, otherwise if not
        # provided it defaults to True.
        is_master: Optional[str] = parse_qs(url.query).get("is_master")
        if is_master:
            pool.is_master = to_bool(is_master[0])

        return pool


def get_connection_factory(
    path: Optional[str] = None, options: Optional[Dict[str, Any]] = None
) -> BaseConnectionFactory:
    """Given a ConnectionFactory class path and options,
    return a ConnectionFactory instance."""
    # TODO read from env.
    default_path: str = "throttled.store.ConnectionFactory"
    cls: Type[BaseConnectionFactory] = import_string(path or default_path)
    return cls(options or {})



import asyncio
import platform
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from importlib import import_module
from typing import Any, Callable, Coroutine, List


def now_sec() -> int:
    return int(time.time())


def now_sec_f() -> float:
    return time.time()


def now_ms() -> int:
    return int(time.time() * 1000)


class Benchmark:
    def __init__(self):
        self.handled_ns_list: List[int] = []
        self.start_times: List[int] = []
        self.end_times: List[int] = []
        self.last_avg: float = 0
        self.last_qps: float = 0

        self._loop = None
        self._has_checked_environment: bool = False

    def __enter__(self):
        self._checked_environment()
        self.clear()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stats()

    async def __aenter__(self):
        self._checked_environment()
        self.clear()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.clear()

    def stats(self):
        total: int = len(self.handled_ns_list)
        avg: float = sum(self.handled_ns_list) / total
        qps: int = int(total / ((max(self.end_times) - min(self.start_times)) / 1e9))

        growth: str = "--"
        growth_rate: float = 0
        if self.last_qps:
            growth_rate: float = (qps - self.last_qps) * 100 / self.last_qps
            growth = f"{('⬆️', '⬇️')[growth_rate < 0]}{growth_rate:.2f}%"

        growth_emo: str = ("🚀", "💤")[growth_rate < 0]
        print(
            f"✅ Total: {total}, "
            f"🕒 Latency: {avg / 1e6:.4f} ms/op, "
            f"{growth_emo}Throughput: {qps} req/s ({growth})"
        )

        self.last_qps = qps
        self.last_avg = avg

    def clear(self):
        self.handled_ns_list.clear()
        self.end_times.clear()
        self.start_times.clear()

    def _checked_environment(self):
        if self._has_checked_environment:
            return

        self._has_checked_environment = True

        print(f"Python {sys.version}")
        print(f"Implementation: {platform.python_implementation()}")
        print(
            f"OS: {platform.system()} {platform.release()}, "
            f"Arch: {platform.machine()} \n"
        )

    def _timer(self, task: Callable[..., Any]) -> Callable[..., Any]:
        def inner(*args, **kwargs):
            start: int = time.perf_counter_ns()
            self.start_times.append(start)
            ret: Any = task(*args, **kwargs)
            end: int = time.perf_counter_ns()
            self.end_times.append(end)
            self.handled_ns_list.append(end - start)
            return ret

        return inner

    def concurrent(
        self, task: Callable[..., Any], batch: int, workers: int = 32, *args, **kwargs
    ) -> List[Any]:
        with self:
            with ThreadPoolExecutor(max_workers=workers) as executor:
                return list(
                    executor.map(
                        lambda _: self._timer(task)(*args, **kwargs), range(batch)
                    )
                )

    def _atimer(self, task: Callable[..., Coroutine]) -> Callable[..., Coroutine]:
        async def inner(*args, **kwargs):
            start: int = time.perf_counter_ns()
            self.start_times.append(start)
            ret = await task(*args, **kwargs)
            end: int = time.perf_counter_ns()
            self.end_times.append(end)
            self.handled_ns_list.append(end - start)
            return ret

        return inner

    async def async_concurrent(
        self,
        task: Callable[..., Coroutine],
        batch: int,
        workers: int = 32,
        *args,
        **kwargs,
    ) -> List[Any]:
        if not self._loop:
            self._loop = asyncio.get_event_loop()

        sem = asyncio.Semaphore(workers)

        async def limited_task():
            async with sem:
                return await self._atimer(task)(*args, **kwargs)

        with self:
            return await asyncio.gather(*[limited_task() for __ in range(batch)])

    def serial(self, task: Callable[..., Any], batch: int, *args, **kwargs) -> List[Any]:
        with self:
            return [self._timer(task)(*args, **kwargs) for __ in range(batch)]


# --------------------------------------------------------------------------------------
# Copyright (c) Django Software Foundation and individual contributors.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of Django nor the names of its contributors may be used
#        to endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


def import_string(dotted_path: str):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        ) from err



from enum import Enum
from typing import List, Optional, Sequence, Tuple, Type

from ..constants import RateLimiterType, StoreType
from ..store import BaseAtomicAction, MemoryStoreBackend, RedisStoreBackend
from ..types import AtomicActionTypeT, KeyT, RateLimiterTypeT, StoreValueT
from ..utils import now_sec
from . import BaseRateLimiter, RateLimitResult, RateLimitState


class FixedWindowAtomicActionType(Enum):
    """Enumeration for types of AtomicActions used in FixedWindowRateLimiter."""

    LIMIT: AtomicActionTypeT = "limit"


class RedisLimitAtomicAction(BaseAtomicAction):
    """Redis-based implementation of AtomicAction for FixedWindowRateLimiter."""

    TYPE: AtomicActionTypeT = FixedWindowAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.REDIS.value

    SCRIPTS: str = """
    local period = tonumber(ARGV[1])
    local limit = tonumber(ARGV[2])
    local cost = tonumber(ARGV[3])
    local current = redis.call("INCRBY", KEYS[1], cost)

    if current == cost then
        redis.call("EXPIRE", KEYS[1], period)
    end

    return {current > limit and 1 or 0, current}
    """

    def __init__(self, backend: RedisStoreBackend):
        # In single command scenario, lua has no performance advantage, and even causes
        # a decrease in performance due to the increase in transmission content.
        # Benchmarks(Python 3.8, Darwin 23.6.0, Arm)
        # >> Redis baseline
        # command    -> set key value
        # serial     -> 🕒Latency: 0.0609 ms/op, 🚀Throughput: 16271 req/s
        # concurrent -> 🕒Latency: 0.4515 ms/op, 💤Throughput: 12100 req/s
        # >> Lua
        # serial     -> 🕒Latency: 0.0805 ms/op, 🚀Throughput: 12319 req/s
        # concurrent -> 🕒Latency: 0.6959 ms/op, 💤Throughput: 10301 req/s
        # >> 👍 Single Command
        # serial     -> 🕒Latency: 0.0659 ms/op, 🚀Throughput: 15040 req/s
        # concurrent -> 🕒Latency: 0.9084 ms/op, 💤Throughput: 11539 req/s
        # self._script: Script = backend.get_client().register_script(self.SCRIPTS)
        self._backend: RedisStoreBackend = backend

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int]:
        period, limit, cost = args[0], args[1], args[2]
        current: int = self._backend.get_client().incrby(keys[0], cost)
        if current == cost:
            self._backend.get_client().expire(keys[0], period)
        return [0, 1][current > limit], current


class MemoryLimitAtomicAction(BaseAtomicAction):
    """Memory-based implementation of AtomicAction for FixedWindowRateLimiter."""

    TYPE: AtomicActionTypeT = FixedWindowAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.MEMORY.value

    def __init__(self, backend: MemoryStoreBackend):
        self._backend: MemoryStoreBackend = backend

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int]:
        key: str = keys[0]
        period: int = args[0]
        limit: int = args[1]
        cost: int = args[2]
        with self._backend.lock:
            current: Optional[int] = self._backend.get(key)
            if current is None:
                current = cost
                self._backend.set(key, current, period)
            else:
                current += cost
                self._backend.get_client()[key] = current

        return (0, 1)[current > limit], current


class FixedWindowRateLimiter(BaseRateLimiter):
    """Concrete implementation of BaseRateLimiter using fixed window as algorithm."""

    class Meta:
        type: RateLimiterTypeT = RateLimiterType.FIXED_WINDOW.value

    @classmethod
    def _default_atomic_action_classes(cls) -> List[Type[BaseAtomicAction]]:
        return [RedisLimitAtomicAction, MemoryLimitAtomicAction]

    @classmethod
    def _supported_atomic_action_types(cls) -> List[AtomicActionTypeT]:
        return [FixedWindowAtomicActionType.LIMIT.value]

    def _prepare(self, key: str) -> Tuple[str, int, int, int]:
        now: int = now_sec()
        period: int = self.quota.get_period_sec()
        period_key: str = f"{key}:period:{now // period}"
        return self._prepare_key(period_key), period, self.quota.get_limit(), now

    def _limit(self, key: str, cost: int = 1) -> RateLimitResult:
        period_key, period, limit, now = self._prepare(key)
        limited, current = self._atomic_actions[
            FixedWindowAtomicActionType.LIMIT.value
        ].do([period_key], [period, limit, cost])

        # |-- now % period --|-- reset_after --|----- next period -----|
        # |--------------- period -------------|
        reset_after: float = period - (now % period)
        return RateLimitResult(
            limited=bool(limited),
            state=RateLimitState(
                limit=limit,
                remaining=max(0, limit - current),
                reset_after=reset_after,
                retry_after=(0, reset_after)[limited],
            ),
        )

    def _peek(self, key: str) -> RateLimitState:
        period_key, period, limit, now = self._prepare(key)
        current: int = int(self._store.get(period_key) or 0)
        return RateLimitState(
            limit=limit,
            remaining=max(0, limit - current),
            reset_after=period - (now % period),
        )
